-- SPDX-FileCopyrightText: 2021 Oxhead Alpha
-- SPDX-License-Identifier: LicenseRef-MIT-OA

-- | Michelson contract in untyped model.

module Morley.Michelson.Untyped.Contract
  ( EntriesOrder (..)
  , Entry(..)
  , canonicalEntriesOrder
  , mapEntriesOrdered
  , mkEntriesOrder

  , ContractBlock (..)
  , orderContractBlock

  , Contract' (..)
  , View' (..)
  , Storage
  , mapContractCode
  ) where

import Control.Lens (folded, makePrisms)
import Data.Aeson (FromJSON, FromJSONKey(..), ToJSON, ToJSONKey(..))
import Data.Aeson.TH (deriveJSON)
import Data.Aeson.Types qualified as AesonTypes
import Data.Data (Data(..))
import Data.Default (Default(..))
import Data.Map qualified as Map
import Data.Text (stripPrefix)
import Fmt (Buildable(build), pretty)
import Text.PrettyPrint.Leijen.Text (nest, semi, text, (<$$>), (<+>))
import Text.PrettyPrint.Leijen.Text qualified as PP

import Morley.Michelson.Printer.Util
  (Prettier(..), RenderDoc(..), assertParensNotNeeded, buildRenderDoc, needsParens, renderOpsList)
import Morley.Michelson.Untyped.Type (ParameterType(..), Ty(..))
import Morley.Michelson.Untyped.View
import Morley.Util.Aeson

data Entry
  = EntryParameter
  | EntryStorage
  | EntryCode
  | EntryView ViewName
  deriving stock (Eq, Ord, Show, Data, Generic)

deriveJSON morleyAesonOptions ''Entry

instance ToJSONKey Entry where
  toJSONKey = AesonTypes.toJSONKeyText $ \case
    EntryParameter -> "parameter"
    EntryStorage -> "storage"
    EntryCode -> "code"
    EntryView name -> "view:" <> unViewName name

instance FromJSONKey Entry where
  fromJSONKey = AesonTypes.FromJSONKeyTextParser $ \case
    "parameter" -> pure EntryParameter
    "storage" -> pure EntryStorage
    "code" -> pure EntryCode
    x | Just name <- stripPrefix "view:" x
      -> either (fail . pretty) (pure . EntryView) $ mkViewName name
    _ -> fail $ "Unexpected Entry value"

instance NFData Entry

-- | Top-level entries order of the contract.
-- This is preserved due to the fact that it affects
-- the output of pretty-printing and serializing contract.
newtype EntriesOrder = EntriesOrder { unEntriesOrder :: Map Entry Word }
  deriving stock (Eq, Ord, Show, Data)
  deriving newtype (NFData, FromJSON, ToJSON)

-- | Helper to construct 'EntriesOrder' from an ordered list of entires.
-- Duplicate entires are ignored.
mkEntriesOrder :: [Entry] -> EntriesOrder
mkEntriesOrder = EntriesOrder . Map.fromListWith const . (`zip` [0..])

instance Default EntriesOrder where
  def = canonicalEntriesOrder

-- | The canonical entries order which is ordered as follow:
-- @parameter@, @storage@, and @code@.
canonicalEntriesOrder :: EntriesOrder
canonicalEntriesOrder = mkEntriesOrder [EntryParameter, EntryStorage, EntryCode]

-- | Contract block, convenient when parsing
data ContractBlock op
  = CBParam ParameterType
  | CBStorage Ty
  | CBCode [op]
  | CBView (View' op)
  deriving stock (Eq, Show, Functor)

makePrisms ''ContractBlock

-- | Construct a contract representation from the contract blocks (i.e. parameters,
-- storage, code blocks, etc.) in arbitrary order.
-- This makes sure that unique blocks like @code@ do not duplicate, and saves the
-- order in the contract so that it can print the contract blocks in the same
-- order it was parsed.
orderContractBlock :: [ContractBlock op] -> Maybe (Contract' op)
orderContractBlock blocks =
  let
    -- we are sure there are no duplicate view names due to the guard below, so
    -- we can construct a Map right away
    contractViews = unsafe . mkViewsSet $ blocks ^.. folded . _CBView
    blockTypes = blocks <&> \case
      CBParam{} -> EntryParameter
      CBStorage{} -> EntryStorage
      CBCode{} -> EntryCode
      CBView View{..} -> EntryView viewName
    entriesOrder = mkEntriesOrder blockTypes
  in do
    guard $ blockTypes == ordNub blockTypes -- must have no duplicates
    contractParameter <- blocks ^? folded . _CBParam
    contractStorage <- blocks ^? folded . _CBStorage
    contractCode <- blocks ^? folded . _CBCode
    pure Contract{..}

instance Buildable (ContractBlock op) where
  build CBParam{} = "parameter"
  build CBStorage{} = "storage"
  build CBCode{} = "code"
  build (CBView View{..}) = "view \"" <> build viewName <> "\""

-- | Map each contract fields by the given function and sort the output
-- based on the 'EntriesOrder'.
mapEntriesOrdered
  :: Contract' op
  -> (ParameterType -> a)
  -> (Storage -> a)
  -> ([op] -> a)
  -> (View' op -> a)
  -> [a]
mapEntriesOrdered Contract{..} fParam fStorage fCode fView = snd <$> sortWith fst elements
  where
    getElemOrder ty = fromMaybe maxBound $ Map.lookup ty $ unEntriesOrder entriesOrder
    elements
      = first getElemOrder
      <$> [ (EntryParameter, fParam contractParameter)
          , (EntryStorage, fStorage contractStorage)
          , (EntryCode, fCode contractCode)]
      <>  (toList contractViews <&> \v@View{..} -> (EntryView viewName, fView v))

-- | Convenience synonym for 'Ty' representing the storage type
type Storage = Ty

-- | General untyped contract representation.
data Contract' op = Contract
  { contractParameter :: ParameterType
    -- ^ Contract parameter type
  , contractStorage :: Storage
    -- ^ Contract storage type
  , contractCode :: [op]
    -- ^ Contract code as a list of operations
  , entriesOrder :: EntriesOrder
    -- ^ Original order of contract blocks, so that we can print them
    -- in the same order they were read
  , contractViews :: ViewsSet op
    -- ^ Contract views
  } deriving stock (Eq, Show, Functor, Data, Generic)

instance NFData op => NFData (Contract' op)

instance (RenderDoc op) => RenderDoc (Contract' op) where
  renderDoc pn contract =
    assertParensNotNeeded pn
      $ foldr (<$$>) (text "")
      $ mapEntriesOrdered contract
        (\parameter -> "parameter" <+> renderDoc needsParens (Prettier parameter) <> semi)
        (\storage -> "storage" <+> renderDoc needsParens (Prettier storage) <> semi)
        (\code -> "code" <+> nest (length ("code {" :: Text)) (renderOpsList False code <> semi))
        (\View{..} -> PP.group $ "view" <+> PP.align (
          PP.sep
            [ renderViewName viewName
            , renderDoc needsParens viewArgument
            , renderDoc needsParens viewReturn
            ]
          PP.<$> renderOpsList False viewCode <> semi
          ))

instance RenderDoc op => Buildable (Contract' op) where
  build = buildRenderDoc

-- | Map all the instructions appearing in the contract.
mapContractCode :: (op -> op) -> Contract' op -> Contract' op
mapContractCode f = fmap f
{-# DEPRECATED mapContractCode "Use fmap instead" #-}

deriveJSON morleyAesonOptions ''Contract'
